/* 
 *	HT Editor
 *	htelf.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htelf.h"
#include "htelfhd.h"
#include "htelfshs.h"
#include "htelfphs.h"
#include "htelfsym.h"
#include "htelfrel.h"
#include "htelfimg.h"
#include "stream.h"

#include "elfstruc.h"

#include <stdlib.h>

format_viewer_if *htelf_ifs[] = {
	&htelfheader_if,
	&htelfsectionheaders_if,
	&htelfprogramheaders_if,
	&htelfimage_if,
	0
};

ht_view *htelf_init(bounds *b, ht_streamfile *file, ht_format_group *format_group)
{
	elf_unsigned_char ident[EI_NIDENT];
	file->seek(0);
	file->read(&ident, sizeof ident);
	if ((ident[ELF_EI_MAG0]!=ELFMAG0) || (ident[ELF_EI_MAG1]!=ELFMAG1) ||
		(ident[ELF_EI_MAG2]!=ELFMAG2) || (ident[ELF_EI_MAG3]!=ELFMAG3) ) return 0;
	ht_elf *g=new ht_elf();
	g->init(b, file, htelf_ifs, format_group, 0);
	return g;
}

format_viewer_if htelf_if = {
	htelf_init,
	0
};

void ht_elf::init(bounds *b, ht_streamfile *file, format_viewer_if **ifs, ht_format_group *format_group, FILEOFS header_ofs)
{
	ht_format_group::init(b, VO_SELECTABLE | VO_BROWSABLE, DESC_ELF, file, false, true, 0, format_group);
	VIEW_DEBUG_NAME("ht_elf");

	ht_elf_shared_data *elf_shared=new ht_elf_shared_data;
	
	shared_data=elf_shared;
	elf_shared->header_ofs=header_ofs;
	elf_shared->shnames=0;
	elf_shared->symtables=0;
	elf_shared->reloctables=0;
	elf_shared->v_image=0;
	
/* read header */
	file->seek(header_ofs);
	file->read(&elf_shared->header, sizeof elf_shared->header);

/* read section headers */
	elf_shared->sheaders.count=elf_shared->header.e_shnum;
	elf_shared->sheaders.sheaders=(ELF_SECTION_HEADER*)malloc(elf_shared->sheaders.count*sizeof *elf_shared->sheaders.sheaders);
	file->seek(header_ofs+elf_shared->header.e_shoff);
	file->read(elf_shared->sheaders.sheaders, elf_shared->sheaders.count*sizeof *elf_shared->sheaders.sheaders);

/* read program headers */
	elf_shared->pheaders.count=elf_shared->header.e_phnum;
	elf_shared->pheaders.pheaders=(ELF_PROGRAM_HEADER*)malloc(elf_shared->pheaders.count*sizeof *elf_shared->pheaders.pheaders);
	file->seek(header_ofs+elf_shared->header.e_phoff);
	file->read(elf_shared->pheaders.pheaders, elf_shared->pheaders.count*sizeof *elf_shared->pheaders.pheaders);
	
/* init ifs */
	ht_format_group::init_ifs(ifs);
	while (init_if(&htelfsymboltable_if)) elf_shared->symtables++;
	while (init_if(&htelfreloctable_if)) elf_shared->reloctables++;
}

void ht_elf::done()
{
	ht_format_group::done();
}

/*
 *	address conversion routines
 */

bool elf_phys_and_mem_section(ELF_SECTION_HEADER *s)
{
	return ((s->sh_type==ELF_SHT_PROGBITS) && (s->sh_addr!=0));
}

bool elf_valid_section(ELF_SECTION_HEADER *s)
{
	return (((s->sh_type==ELF_SHT_PROGBITS) || (s->sh_type==ELF_SHT_NOBITS)) && (s->sh_addr!=0));
}

bool elf_addr_to_ofs(elf_section_headers *section_headers, ADDR addr, dword *ofs)
{
	ELF_SECTION_HEADER *s=section_headers->sheaders;
	for (UINT i=0; i<section_headers->count; i++) {
		if ((elf_phys_and_mem_section(s)) && (addr>=s->sh_addr) && (addr<s->sh_addr+s->sh_size)) {
			*ofs=addr-s->sh_addr+s->sh_offset;
			return 1;
		}
		s++;
	}
	return 0;
}

bool elf_addr_to_section(elf_section_headers *section_headers, ADDR addr, int *section)
{
	ELF_SECTION_HEADER *s=section_headers->sheaders;
	for (UINT i=0; i<section_headers->count; i++) {
		if ((elf_valid_section(s)) && (addr>=s->sh_addr) && (addr<s->sh_addr+s->sh_size)) {
			*section=i;
			return 1;
		}
		s++;
	}
	return 0;
}

bool elf_addr_is_valid(elf_section_headers *section_headers, ADDR addr)
{
	ELF_SECTION_HEADER *s=section_headers->sheaders;
	for (UINT i=0; i<section_headers->count; i++) {
		if ((elf_valid_section(s)) && (addr>=s->sh_addr) && (addr<s->sh_addr+s->sh_size)) {
			return 1;
		}
		s++;
	}
	return 0;
}

bool elf_addr_is_physical(elf_section_headers *section_headers, ADDR addr)
{
	return 0;
}


/*
 *	offset conversion routines
 */

bool elf_ofs_to_addr(elf_section_headers *section_headers, dword ofs, ADDR *addr)
{
	ELF_SECTION_HEADER *s=section_headers->sheaders;
	for (UINT i=0; i<section_headers->count; i++) {
		if ((elf_phys_and_mem_section(s)) && (ofs>=s->sh_offset) && (ofs<s->sh_offset+s->sh_size)) {
			*addr=ofs-s->sh_offset+s->sh_addr;
			return 1;
		}
		s++;
	}
	return 0;
}

bool elf_ofs_to_section(elf_section_headers *section_headers, dword ofs, ADDR *section)
{
	ELF_SECTION_HEADER *s=section_headers->sheaders;
	for (UINT i=0; i<section_headers->count; i++) {
		if ((elf_valid_section(s)) && (ofs>=s->sh_offset) && (ofs<s->sh_offset+s->sh_size)) {
			*section=i;
			return 1;
		}
		s++;
	}
	return 0;
}

bool elf_ofs_to_addr_and_section(elf_section_headers *section_headers, dword ofs, ADDR *addr, ADDR *section)
{
	return 0;
}
