/*
 *	HT Editor
 *	main.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>

#include "analy_register.h"
#include "htapp.h"
#include "htatom.h"
#include "htcfg.h"
#include "htclipboard.h"
#include "htcurses.h"
#include "hthelp.h"
#include "hthist.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htidle.h"
#include "htmenu.h"
#include "htpal.h"
#include "htinfo.h"
#include "htreg.h"
#include "htsys.h"
#include "stddata.h"

char *htcopyrights[]=
{
	ht_loadtitle,
	htanalyser_loadtitle,
	0
};

/*void copyright()
{
	char **copyrights=htcopyrights;
	while (*copyrights) {
		printf(*copyrights);
		printf("\n");
		copyrights++;
	}
	printf(htlicenseinfo);
}*/

typedef bool (*initfunc)();
typedef void (*donefunc)();

struct initdonefunc {
	initfunc i;
	donefunc d;
	char *name;
};

#define INITDONE(name) { init_##name, done_##name, #name }

initdonefunc initdone[] = {
	INITDONE(atom),
	INITDONE(registry),
	INITDONE(pal),
	INITDONE(keyb),
	INITDONE(idle),
	INITDONE(menu),
	INITDONE(hist),
	INITDONE(app),
	INITDONE(help),
	INITDONE(clipboard),
	INITDONE(data),
	INITDONE(obj),
	INITDONE(string),
	INITDONE(analyser),
	INITDONE(stddata),
	INITDONE(cfg)
};

int htstate;

bool init()
{
	for (htstate=0; htstate<(int)(sizeof initdone / sizeof initdone[0]); htstate++) {
		if (!initdone[htstate].i()) return false;
	}
	return true;
}

void done()
{
	for (htstate--; htstate>=0; htstate--) {
		initdone[htstate].d();
	}
}

int main(int argc, char *argv[])
{
	this_app=argv[0];

	if (!init()) {
		int init_failed = htstate;
		done();
		printf("init(): fatal error in init_%s\n", initdone[init_failed].name);
		return 1;
	}

	((ht_app*)app)->sendmsg(msg_draw);

	char **copyrights=htcopyrights;
	while (*copyrights) {
		LOG(*copyrights);
		copyrights++;
	}

	bool overwrite_config = true;
	loadstore_result load=load_projectconfig();
	switch (load) {
		case LS_OK:
			break;
		case LS_ERROR_NOT_FOUND:
			LOG("couldn't load configuration file...");
			break;
		case LS_ERROR_READ:
			LOG("couldn't read configuration file...");
			infobox("couldn't read configuration file...");
			break;
		case LS_ERROR_MAGIC:
			LOG("configuration file has wrong magic...");
			if (confirmbox_modal("%s\noverwrite?", "configuration file has wrong magic...")!=button_ok) {
				overwrite_config = false;
			}
			break;
		case LS_ERROR_VERSION:
			LOG("configuration file has wrong version...");
			if (confirmbox_modal("%s\noverwrite?", "configuration file has wrong version...")!=button_ok) {
				overwrite_config = false;
			}
			break;
		case LS_ERROR_CORRUPTED:
//			done();
			printf("fatal error loading configuration file ("PROJECT_CONFIG_FILE_NAME").\nplease try to delete it.\n");
			return 1;
		default: {assert(0);}
	}

	int filecount=0;
	for (int i=1; i<argc; i++) {
		if (argv[i][0]=='-') {
			LOG("unrecognized option: %s", argv[i]);
		} else {
			if (((ht_app*)app)->create_file_window(argv[i])) {
				filecount++;
			}
		}
	}

	((ht_app*)app)->run(0);

	loadstore_result save=LS_OK;
	if (overwrite_config) {
		LOG("saving config...");
		save=save_projectconfig();
	}
	LOG("exit.");
	done();
	if (save!=LS_OK) {
		printf("save_projectfile(): error\n");
		return 1;
	}
	return 0;
}

